package handler

import (
	"context"
	"errors"
	"io"
	"net/http"
	"net/url"
	"strconv"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/tree"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/utils"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"github.com/valyala/fasthttp"
	"go.uber.org/zap"
)

// max bytes needed to detect content type according to http.DetectContentType docs.
const sizeToDetectType = 512

const (
	hdrObjectID    = "X-Object-Id"
	hdrOwnerID     = "X-Owner-Id"
	hdrContainerID = "X-Container-Id"
)

func (h *Handler) headObject(ctx context.Context, req *fasthttp.RequestCtx, objectAddress oid.Address) {
	var start = time.Now()

	btoken := bearerToken(ctx)

	prm := PrmObjectHead{
		PrmAuth: PrmAuth{
			BearerToken: btoken,
		},
		Address: objectAddress,
	}

	obj, err := h.frostfs.HeadObject(ctx, prm)
	if err != nil {
		h.logAndSendError(ctx, req, logs.FailedToHeadObject, err, zap.Stringer("elapsed", time.Since(start)))
		return
	}

	req.Response.Header.Set(fasthttp.HeaderContentLength, strconv.FormatUint(obj.PayloadSize(), 10))
	var (
		contentType string
		filename    string
		filepath    string
	)
	for _, attr := range obj.Attributes() {
		key := attr.Key()
		val := attr.Value()
		if !isValidToken(key) || !isValidValue(val) {
			continue
		}

		key = utils.BackwardTransformIfSystem(key)

		req.Response.Header.Set(utils.UserAttributeHeaderPrefix+key, val)
		switch key {
		case object.AttributeTimestamp:
			value, err := strconv.ParseInt(val, 10, 64)
			if err != nil {
				h.reqLogger(ctx).Info(logs.CouldntParseCreationDate,
					zap.String("key", key),
					zap.String("val", val),
					zap.Error(err),
					logs.TagField(logs.TagDatapath))
				continue
			}
			req.Response.Header.Set(fasthttp.HeaderLastModified, time.Unix(value, 0).UTC().Format(http.TimeFormat))
		case object.AttributeContentType:
			contentType = val
		case object.AttributeFilePath:
			filepath = val
		case object.AttributeFileName:
			filename = val
		}
	}
	if filename == "" {
		filename = filepath
	}

	idsToResponse(&req.Response, obj)

	if len(contentType) == 0 {
		contentType, _, err = readContentType(obj.PayloadSize(), func(sz uint64) (io.Reader, error) {
			prmRange := PrmObjectRange{
				PrmAuth: PrmAuth{
					BearerToken: btoken,
				},
				Address:      objectAddress,
				PayloadRange: [2]uint64{0, sz},
			}

			return h.frostfs.RangeObject(ctx, prmRange)
		}, filename)
		if err != nil && err != io.EOF {
			h.logAndSendError(ctx, req, logs.FailedToDetectContentTypeFromPayload, err, zap.Stringer("elapsed", time.Since(start)))
			return
		}
	}
	req.SetContentType(contentType)
}

func idsToResponse(resp *fasthttp.Response, obj *object.Object) {
	objID, _ := obj.ID()
	cnrID, _ := obj.ContainerID()
	resp.Header.Set(hdrObjectID, objID.String())
	resp.Header.Set(hdrOwnerID, obj.OwnerID().String())
	resp.Header.Set(hdrContainerID, cnrID.String())
}

// HeadByAddressOrBucketName handles head requests using simple cid/oid or bucketname/key format.
func (h *Handler) HeadByAddressOrBucketName(req *fasthttp.RequestCtx) {
	ctx, span := tracing.StartSpanFromContext(utils.GetContextFromRequest(req), "handler.HeadByAddressOrBucketName")
	defer span.End()

	cidParam, _ := req.UserValue("cid").(string)
	oidParam, _ := req.UserValue("oid").(string)

	ctx = utils.SetReqLog(ctx, h.reqLogger(ctx).With(
		zap.String("cid", cidParam),
		zap.String("oid", oidParam),
	))

	path, err := url.QueryUnescape(oidParam)
	if err != nil {
		h.logAndSendError(ctx, req, logs.FailedToUnescapePath, err)
		return
	}

	bktInfo, err := h.getBucketInfo(ctx, cidParam)
	if err != nil {
		h.logAndSendError(ctx, req, logs.FailedToGetBucketInfo, err)
		return
	}

	checkS3Err := h.tree.CheckSettingsNodeExists(ctx, bktInfo)
	if checkS3Err != nil && !errors.Is(checkS3Err, tree.ErrNodeNotFound) {
		h.logAndSendError(ctx, req, logs.FailedToCheckIfSettingsNodeExist, checkS3Err)
		return
	}

	prm := MiddlewareParam{
		Context: ctx,
		Request: req,
		BktInfo: bktInfo,
		Path:    path,
	}

	indexPageEnabled := h.config.IndexPageEnabled()

	if checkS3Err == nil {
		run(prm, h.errorMiddleware(logs.ObjectNotFound, tree.ErrNodeNotFound),
			Middleware{Func: h.byS3PathMiddleware(h.headObject, noopFormer), Enabled: true},
			Middleware{Func: h.byS3PathMiddleware(h.headObject, indexFormer), Enabled: indexPageEnabled},
		)
	} else {
		slashFallbackEnabled := h.config.EnableFilepathSlashFallback()
		fileNameFallbackEnabled := h.config.EnableFilepathFallback()

		run(prm, h.errorMiddleware(logs.ObjectNotFound, ErrObjectNotFound),
			Middleware{Func: h.byAddressMiddleware(h.headObject), Enabled: true},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFilePath, noopFormer), Enabled: true},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFilePath, reverseLeadingSlash), Enabled: slashFallbackEnabled},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFileName, noopFormer), Enabled: fileNameFallbackEnabled},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFileName, reverseLeadingSlash), Enabled: fileNameFallbackEnabled && slashFallbackEnabled},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFilePath, indexFormer), Enabled: indexPageEnabled},
			Middleware{Func: h.byAttributeSearchMiddleware(h.headObject, object.AttributeFileName, indexFormer), Enabled: fileNameFallbackEnabled && indexPageEnabled},
		)
	}
}

// HeadByAttribute handles attribute-based head requests.
func (h *Handler) HeadByAttribute(req *fasthttp.RequestCtx) {
	ctx, span := tracing.StartSpanFromContext(utils.GetContextFromRequest(req), "handler.HeadByAttribute")
	defer span.End()

	h.byAttribute(ctx, req, h.headObject)
}

func (h *Handler) errorMiddleware(msg string, err error) MiddlewareFunc {
	return func(prm MiddlewareParam) bool {
		h.logAndSendError(prm.Context, prm.Request, msg, err)
		return false
	}
}
